#! /bin/bash
#
# Script to generate the version info of GEMMA and its environment
# in ./src/version.h

DATE=$(date "+%Y-%m-%d")
YEAR=$(date "+%Y")
PROFILE=$1

echo // version.h generated by GEMMA $0
echo \#define GEMMA_VERSION \"$(cat ./VERSION)\"
echo \#define GEMMA_DATE \"$DATE\"
echo \#define GEMMA_YEAR \"$YEAR\"
echo \#define GEMMA_PROFILE \"$PROFILE\"
